/*
** Beaver's an Early AdVanced EditoR
** (C) 1999-2000 Marc Bevand, Damien Terrier and Emmanuel Turquin
**
** conf.c
**
** Author<s>:   Damien Terrier   (aka "mAd")   on vacation...
** Last update: Sat Jun  3 00:50:01 2000
** Description: Beaver preferences management source
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "struct.h"
#include "conf.h"


/**************************** PUBLIC FUNCTIONS *******************************/


/* Allow you to record a string in a section in a file.
   For instance, 'set_string_conf ("france/paris/epita", "toto");' puts:
   [paris]
   epita = toto
   in the file '$HOME/.AppName/france' */

void set_string_conf (gchar *KeyToSet, gchar *ValueToSet)
{
  gchar **SplittedValue;
  t_content Content;
  t_section_limits SectionLimits;
  gchar *ValuePath;
  
  SplittedValue = g_strsplit (KeyToSet, "/", 0);
  ValuePath = g_strconcat (g_get_home_dir(), PATH_SEP_STRING, CONF_DIR,
			   PATH_SEP_STRING, SplittedValue[0], NULL);
  Content = get_file_content (ValuePath);
  SectionLimits = get_section_limits (Content, ValuePath, SplittedValue[1]);
  Content = get_file_content (ValuePath);
  Content = set_string_value (SectionLimits, SplittedValue[2],
			      Content, ValueToSet);
  set_file_content (ValuePath, Content);
  g_free (ValuePath);
}


/* Allow you to record a value (INT or  BOOLEAN) in a section in a file.
   For instance, 'set_conf ("france/paris/epita", "INT", 42);' puts:
   [paris]
   epita = 42
   in the file '$HOME/.AppName/france' */

void set_conf (gchar *KeyToSet, gchar* ValueType, gint ValueToSet)
{
  gchar **SplittedValue;
  t_content Content;
  t_section_limits SectionLimits;
  gchar *ValuePath;
  
  SplittedValue = g_strsplit (KeyToSet, "/", 0);
  ValuePath = g_strconcat (g_get_home_dir(), PATH_SEP_STRING, CONF_DIR,
			   PATH_SEP_STRING, SplittedValue[0], NULL);
  Content = get_file_content (ValuePath);
  SectionLimits = get_section_limits (Content, ValuePath, SplittedValue[1]);
  Content = get_file_content (ValuePath);
  if (!g_strcasecmp (ValueType, "INT"))
    Content = set_value (SectionLimits, SplittedValue[2],
			 Content, "INT", ValueToSet);
  else if (!g_strcasecmp (ValueType, "BOOLEAN"))
    Content = set_value (SectionLimits, SplittedValue[2],
			 Content, "BOOLEAN", ValueToSet);
  set_file_content (ValuePath, Content);
  g_free (ValuePath);
}


/* The 3 following functions use the same syntax as the two above.
   For example, 'get_int_conf ("france/paris/epita");' returns 42, given
   that the file '$HOME/.AppName/france' contains:
   (...)
   [paris]
   epita = 42
   (...) */

gint get_int_conf (gchar *ValueToFind)
{
  gchar **SplittedValue;
  t_content Content;
  t_section_limits SectionLimits;
  gchar *ValuePath;

  SectionLimits.Begin = 0;
  SectionLimits.End = 0;
  Content.LineNumber = 0;
  ValueToFind = g_strconcat (ValueToFind, " = ", NULL);
  SplittedValue = g_strsplit (ValueToFind, "/", 0);
  ValuePath = g_strconcat (g_get_home_dir(), PATH_SEP_STRING, CONF_DIR,
			   PATH_SEP_STRING, SplittedValue[0], NULL);
  Content = get_file_content (ValuePath);
  SectionLimits = get_section_limits (Content, ValuePath, SplittedValue[1]);
  g_free (ValuePath);
  return (get_int_value (SectionLimits, SplittedValue[2], Content));
}


gboolean get_bool_conf (gchar *ValueToFind)
{
  gchar **SplittedValue;
  t_content Content;
  t_section_limits SectionLimits;
  gchar *ValuePath;

  SectionLimits.Begin = 0;
  SectionLimits.End = 0;
  Content.LineNumber = 0;
  ValueToFind = g_strconcat (ValueToFind, " = ", NULL);
  SplittedValue = g_strsplit (ValueToFind, "/", 0);
  ValuePath = g_strconcat (g_get_home_dir(), PATH_SEP_STRING, CONF_DIR,
			   PATH_SEP_STRING, SplittedValue[0], NULL);
  Content = get_file_content (ValuePath);
  SectionLimits = get_section_limits (Content, ValuePath, SplittedValue[1]);
  g_free (ValuePath);
  return (get_bool_value (SectionLimits, SplittedValue[2], Content));
}


gchar *get_string_conf (gchar *ValueToFind)
{
  gchar **SplittedValue;
  t_content Content;
  t_section_limits SectionLimits;
  gchar *ValuePath;

  SectionLimits.Begin = 0;
  SectionLimits.End = 0;
  Content.LineNumber = 0;
  ValueToFind = g_strconcat (ValueToFind, " = ", NULL);
  SplittedValue = g_strsplit (ValueToFind, "/", 0);
  ValuePath = g_strconcat (g_get_home_dir(), PATH_SEP_STRING, CONF_DIR,
			   PATH_SEP_STRING, SplittedValue[0], NULL);
  Content = get_file_content (ValuePath);
  SectionLimits = get_section_limits (Content, ValuePath, SplittedValue[1]);
  g_free (ValuePath);
  return (get_string_value (SectionLimits, SplittedValue[2], Content));
}


/************************ AUTHORIZED PERSONAL ONLY ***************************/


/* Put the content of a file in a table */

t_content get_file_content (gchar *OpenedFile)
{
  FILE *File;
  t_content Content;
  gchar Char;
  gint i;
  gint j;
  
  Content.LineNumber = 0;
  for (i = 0; i < PREFS_MAX_LINE; i++) Content.Text[i][0] = '\0';
  if ((File = fopen (OpenedFile, "r")) != NULL)
    {
      i = 0;
      j = 0;
      while ((Char = getc(File)) != EOF)
	{
	  if (Char == '\n')
	    {
	      Content.LineNumber ++;
	      Content.Text[i][j] = '\n';
	      Content.Text[i][j+1] = '\0';
	      i++;
	      j = 0;
	    }
	  else
	    Content.Text[i][j++] = Char;
	}
      fclose (File);
    }
  return (Content);
}


/* Set a file content from a given table */

void set_file_content (gchar *FileName, t_content Content)
{
  FILE *File;
  gint i ;
  
  File = fopen (FileName, "w");
  for (i = 0; i < Content.LineNumber; i++)
    fprintf (File, "%s", Content.Text[i]);
  fclose (File);
}


/* Give the limits of a section (a section is represented by '[section]' */

t_section_limits get_section_limits (t_content Content, gchar *FileName,
				     gchar *Section)
{
  gint i = 0;
  t_section_limits SectionLimits;
  
  SectionLimits.Begin = 0;
  SectionLimits.End = 0;
  while (strncasecmp (g_strconcat("[", Section, "]", NULL) ,
		      Content.Text[i], (strlen (Section)+2)))
    if (i >= Content.LineNumber)
      {
	Content.LineNumber += 2;
	strcpy (Content.Text[i], g_strconcat("[", Section, "]\n", NULL));
	strcpy (Content.Text[i+1], "\n\0");
	SectionLimits.Begin = i+1;
	SectionLimits.End = i+1;
	set_file_content (FileName, Content);
	return (SectionLimits);
      }
    else i++;
  SectionLimits.Begin = i+1;
  while ((Content.Text[++i][0] != '[') && i < Content.LineNumber);
  SectionLimits.End = i-1 ;
  return (SectionLimits);
}


/* Used by 'set_string_conf' */

t_content set_string_value (t_section_limits SectionLimits,
			    gchar *KeyToSet, t_content Content,
			    gchar *ValueToSet)
{
  gint i, j;
  gchar *Buffer = "";
  
  KeyToSet = g_strconcat (KeyToSet, " = ", NULL);
  Buffer = g_strconcat (KeyToSet, ValueToSet, "\n", NULL);
  for (i = SectionLimits.Begin; i <= SectionLimits.End; i++)
    {
      if (!strncasecmp (Content.Text[i], KeyToSet, strlen (KeyToSet)))
	{
	  strcpy (Content.Text[i], Buffer);
	  break;
	}
    }
  if (i > SectionLimits.End)
    {
      for (j = Content.LineNumber-1; j >= SectionLimits.Begin; j--)
	strcpy (Content.Text[j+1], Content.Text[j]);
      Content.LineNumber ++;
      strcpy (Content.Text[SectionLimits.Begin], Buffer);
    }
  g_free (Buffer);
  return (Content);
}


/* Used by 'set_conf' */

t_content set_value (t_section_limits SectionLimits,
		     gchar *KeyToSet, t_content Content,
		     gchar *ValueType, gint ValueToSet)
{
  gint i, j;
  gchar *Buffer = "";
  
  KeyToSet = g_strconcat (KeyToSet, " = ", NULL);
  if (!g_strcasecmp (ValueType, "INT"))
    Buffer = g_strconcat (KeyToSet, g_strdup_printf ("%d", ValueToSet),
			  "\n", NULL);
  else if (!g_strcasecmp (ValueType, "BOOLEAN"))
    {
      Buffer = ValueToSet? "TRUE" : "FALSE";
      Buffer = g_strconcat (KeyToSet, Buffer, "\n", NULL);
    }
  else if (!g_strcasecmp (ValueType, "STRING"))
    Buffer = g_strconcat (KeyToSet, ValueToSet, "\n", NULL);
  for (i = SectionLimits.Begin; i <= SectionLimits.End; i++)
    {
      if (!strncasecmp (Content.Text[i], KeyToSet, strlen (KeyToSet)))
	{
	  strcpy (Content.Text[i], Buffer);
	  break;
	}
    }
  if (i > SectionLimits.End)
    {
      for (j = Content.LineNumber-1; j >= SectionLimits.Begin; j--)
	strcpy (Content.Text[j+1], Content.Text[j]);
      Content.LineNumber ++;
      strcpy (Content.Text[SectionLimits.Begin], Buffer);
    }
  g_free (Buffer);
  return (Content);
}


/* The 3 functions below are used by 'get_blablabla_value' */

gint get_int_value(t_section_limits SectionLimits,
		   gchar *ValueToFind, t_content Content)
{
  gint i, j, CurrentLine;
  gchar Buffer[PREFS_MAX_LINE_LENGTH] = "";
  
  for (i = SectionLimits.Begin; i <= SectionLimits.End;	i++)
    if (!strncasecmp(Content.Text[i], ValueToFind, strlen (ValueToFind)))
      break;
  if (i > SectionLimits.End) return (-1);
  CurrentLine = i;
  i = strlen (ValueToFind);
  j = 0;
  while (Content.Text[CurrentLine][i] != '\n')
    Buffer[j++] = Content.Text[CurrentLine][i++];
  g_strstrip (Buffer);
  if (!strcmp (Buffer, "")) return (-1);
  return ((gint) (g_strtod (Buffer, NULL)));
}


gboolean get_bool_value(t_section_limits SectionLimits,
			gchar *ValueToFind, t_content Content)
{
  gint i, j, CurrentLine;
  gchar Buffer[PREFS_MAX_LINE_LENGTH] = "";
  
  for (i = SectionLimits.Begin; i <= SectionLimits.End;	i++)
    if (!strncasecmp(Content.Text[i], ValueToFind, strlen (ValueToFind)))
      break;
  if (i > SectionLimits.End) return (FALSE);
  CurrentLine = i;
  i = strlen (ValueToFind);
  j = 0;
  while (Content.Text[CurrentLine][i] != '\n')
    Buffer[j++] = Content.Text[CurrentLine][i++];
  g_strstrip (Buffer);
  return ((gboolean) (!strncasecmp (Buffer, "TRUE", 4)));
}


gchar *get_string_value(t_section_limits SectionLimits,
			gchar *ValueToFind, t_content Content)
{
  gint i, j, CurrentLine;
  gchar Buffer[PREFS_MAX_LINE_LENGTH] = "";
  gchar *RetValue;
  
  for (i = SectionLimits.Begin; i <= SectionLimits.End;	i++)
    if (!strncasecmp(Content.Text[i], ValueToFind, strlen (ValueToFind)))
      break;
  if (i > SectionLimits.End) return ("");
  CurrentLine = i;
  i = strlen (ValueToFind);
  j = 0;
  while (Content.Text[CurrentLine][i] != '\n')
    Buffer[j++] = Content.Text[CurrentLine][i++];
  Buffer[j+1] = '\0';
  g_strstrip (Buffer);
  RetValue = g_strdup (Buffer);
  return (RetValue);
}
