/*
** Beaver's an Early AdVanced EditoR
** (C) 1999-2000 Marc Bevand, Damien Terrier and Emmanuel Turquin
**
** editor.c
**
** Author<s>:   Marc Bevand (aka "After") <bevand_m@epita.fr>
** Last update: Wed Jan 10 20:30:55 2001
** Description: This source stands for text management in beaver (highlighting,
**   auto-indentation)
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
** NOTE_REMARK : syhi stands for SYntax HIghlighting, the main work to
**  be accomplished by the source
*/

/*
** In order to view debugging messages, define some of these csts
*/
//#define DEBUG_AUTO_INDENT
//#define DEBUG_FCN
//#define DEBUG_WORDFILE
//#define DEBUG_EXT_REC
//#define DEBUG_SYHI
//#define DEBUG_CORRECT
//#define DEBUG_WTREE

#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "editor.h"
#include "struct.h"
#include "msgbar.h"
#include "tools.h"
#include "undoredo.h"
#include "conf.h" // Added by Ender

gchar			char_alpha[256] =
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,
 0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

/*
** To access WidgetInfo, we need some external variables, and we do :
** FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
**        WidgetInfo.<an_element_of_structure>)
*/
extern GtkWidget	*MainNotebook;
extern GArray		*FileProperties;

/*
** Some internal global constants
*/
static gint		refresh_from = 0;
static gboolean		enter_pressed = FALSE;
static gint		word_to_correct = -1;

/*
** Fcn called by external source to load content of a file in the widget
**
** Parameters :
**  Editor		The GtkWidget that display text
**  Filename		The file to be displayed in the widget
**
** Return values :
**  -1			Could not allocate color
**  0			All is right and type of file reconized
**  1			All is right but type of file not recognized
*/
extern gint		open_file_in_editor(GtkWidget *Editor,
					    const gchar *Filename)
{
  FILE			*fp;
  gint			filesize;
  gchar			*buffer;
  gint			Lg;
  gint			i, j;
  gboolean		color_alloc_error;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  color_alloc_error = FALSE;
  Prefs.ColMap = gdk_colormap_get_system();
  for (i = 0; (i < MAX_LANG) && !color_alloc_error; i++)
    {
      if (!gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].Comment)) ||
	  !gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].CommentAlt)) ||
	  !gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].String0)) ||
	  !gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].String1)) ||
	  !gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].Number)))
	color_alloc_error = TRUE;
      for (j = 0; (j < MAX_COL) && !color_alloc_error; j++)
	if (!gdk_color_alloc(Prefs.ColMap, &(Prefs.Colors.L[i].C[j])))
	  color_alloc_error = TRUE;
    }
  if (color_alloc_error)
    {
      g_print(__FILE__": "__func__"(): End, color allocation failure\n");
      return -1;
    }
  if ((fp = fopen(Filename, "rt")))
    {
      filesize = 0;
      while (getc(fp) != EOF)
	filesize++;
      if (filesize > 0)
	{
#ifdef DEBUG_FCN
	  g_print(__FILE__": "__func__"(): Filesize > 0, mallocing...\n");
#endif
	  buffer = g_malloc(filesize + 2);
	  fseek(fp, 0, SEEK_SET);
	  fread(buffer, filesize, 1, fp);
	  fclose(fp);
	  buffer[filesize] = '\n';
	  buffer[filesize + 1] = '\0';
#ifdef DEBUG_FCN
	  g_print(__FILE__": "__func__"(): buffer has been fill in\n");
#endif
	}
      else
	buffer = NULL;
    }
  else /* Can not open file */
    {
      buffer = NULL;
      filesize = 0;
    }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Initializing variables...\n");
#endif
  // Modified by Ender : Begin
  Prefs.SyhiDepth = get_int_conf ("General/Adv/SynHighDepth");
  Prefs.AutoCorrection = get_bool_conf ("General/Adv/AutoCorrect");
  Prefs.AutoIndentation = get_bool_conf ("General/Adv/AutoIndent");
  // Modified by Ender : End
  Prefs.TabSize = 8; /* NOTE_WARNING: Not yet implemented :) */
  for (i = 0; i < MAX_LANG; i++)
    {
      Prefs.Colors.L[i].Comment.red   = 0xd000;
      Prefs.Colors.L[i].Comment.green = 0x0000;
      Prefs.Colors.L[i].Comment.blue  = 0x0000;
      Prefs.Colors.L[i].CommentAlt.red   = 0xd000;
      Prefs.Colors.L[i].CommentAlt.green = 0x0000;
      Prefs.Colors.L[i].CommentAlt.blue  = 0x8000;
      Prefs.Colors.L[i].String0.red   = 0xd000; /* String0 is usually "str" */
      Prefs.Colors.L[i].String0.green = 0xb000;
      Prefs.Colors.L[i].String0.blue  = 0x3000;
      Prefs.Colors.L[i].String1.red   = 0x8000; /* String1 is usually 'a' */
      Prefs.Colors.L[i].String1.green = 0x4000;
      Prefs.Colors.L[i].String1.blue  = 0x8000;
      Prefs.Colors.L[i].Number.red   = 0x0000;
      Prefs.Colors.L[i].Number.green = 0x9000;
      Prefs.Colors.L[i].Number.blue  = 0x0000;
      Prefs.Colors.L[i].C[0].red   = 0x0000;
      Prefs.Colors.L[i].C[0].green = 0x4000;
      Prefs.Colors.L[i].C[0].blue  = 0xa000;
      Prefs.Colors.L[i].C[1].red   = 0x0000;
      Prefs.Colors.L[i].C[1].green = 0x8000;
      Prefs.Colors.L[i].C[1].blue  = 0x4000;
      Prefs.Colors.L[i].C[2].red   = 0xd000;
      Prefs.Colors.L[i].C[2].green = 0x5000;
      Prefs.Colors.L[i].C[2].blue  = 0x7000;
      Prefs.Colors.L[i].C[3].red   = 0xe000;
      Prefs.Colors.L[i].C[3].green = 0x9000;
      Prefs.Colors.L[i].C[3].blue  = 0x0000;
      Prefs.Colors.L[i].C[4].red   = 0x8000;
      Prefs.Colors.L[i].C[4].green = 0x8000;
      Prefs.Colors.L[i].C[4].blue  = 0x8000;
      Prefs.Colors.L[i].C[5].red   = 0x8000;
      Prefs.Colors.L[i].C[5].green = 0x8000;
      Prefs.Colors.L[i].C[5].blue  = 0x8000;
      Prefs.Colors.L[i].C[6].red   = 0x8000;
      Prefs.Colors.L[i].C[6].green = 0x8000;
      Prefs.Colors.L[i].C[6].blue  = 0x8000;
      Prefs.Colors.L[i].C[7].red   = 0x8000;
      Prefs.Colors.L[i].C[7].green = 0x8000;
      Prefs.Colors.L[i].C[7].blue  = 0x8000;
      /* And so on: Prefs.Colors.L[i].C[j]. where j is from 0 to MAX_LANG-1 */
    }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End initializing variables\n");
#endif
  /* End, now detecting extension... */
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.Lg) = Lg = guess_lang();
#ifdef DEBUG_EXT_REC
  if (Lg > -1)
    g_print(__FILE__": "__func__"(): Extension recognized: language = /L%i "
	    "\"%s\"\n", Lg + 1, Prefs.L[Lg].Description);
  else
    g_print(__FILE__": "__func__"(): Extension not recognized\n");
#endif
  /* Now, language of buffer = Lg is set */
  if (filesize > 0)
    {
      gtk_text_insert(GTK_TEXT(Editor), NULL, NULL, NULL,
		      buffer, filesize);
#ifdef DEBUG_SYHI
      g_print(__FILE__": "__func__"(): Freeing buffer...\n");
#endif
      g_free(buffer);
#ifdef DEBUG_SYHI
      g_print(__FILE__": "__func__"(): Buffer freed\n");
#endif
    }
  /* NOTE_WARNING: This fcn does not work !!!:
     gtk_text_set_point(GTK_TEXT(Editor), 0); */
  gtk_editable_set_position(GTK_EDITABLE(Editor), 0);
  init_undoredo ();
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectIns) =
    gtk_signal_connect(GTK_OBJECT(Editor), "insert_text",
		       (GtkSignalFunc)text_has_been_inserted,
		       NULL);
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectDel) =
    gtk_signal_connect(GTK_OBJECT(Editor), "delete_text",
		       (GtkSignalFunc)text_has_been_deleted,
		       NULL);
  if (Lg == -1)
    {
#ifdef DEBUG_SYHI
      g_print(__FILE__": "__func__"(): End, language not reconized\n");
#endif
      return 1;
    }
  //NOTE_WARNING: the callback fcn refresh_syhi() needs the "changed"
  //signal to be produced after "delete_text" ans "insert_text" signals 
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectChg) =
    gtk_signal_connect(GTK_OBJECT(Editor), "changed",
		       (GtkSignalFunc)refresh_syhi,
		       (gpointer)0xeeeeeeee);
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectChgAft) =
    gtk_signal_connect_after(GTK_OBJECT(Editor), "changed",
			     (GtkSignalFunc)reconnect_sig,
			     NULL);
  refresh_editor(Editor);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return 0;
}

/*
** Fcn called by external source to reload content of a file in the widget
**
** Parameters :
**  Editor		The GtkWidget that display text
**  Filename		The file to be displayed in the widget
**
** Return values :
**  -1			Could not allocate color
**  0			All is right and type of file reconized
**  1			All is right but type of file not recognized
*/
extern gint		reopen_file_in_editor(GtkWidget *Editor,
					      const gchar *Filename)
{
  gint			i;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectIns));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectDel));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectChg));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectChgAft));
  gtk_editable_delete_text(GTK_EDITABLE(Editor), 0, -1);
  i = open_file_in_editor(Editor, Filename);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return i;
}

/*
** This fcn produce a total refresh of the syhi of the widget
** passed in argument
**
** NOTE_WARNING: Can be called only if open_file_in_editor has been
** called with this widget. If name of file has changed, so
** open_file_in_editor must be called before refresh_editor.
**
** Parameters :
**  Editor		Pointer to the GtkWidget to refresh
**
** Return Values :
**  void
*/
extern void		refresh_editor(GtkWidget *Editor)
{
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  if ((FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	      WidgetInfo.Lg) = guess_lang()) > -1)
    {
      gint		tmp;

      tmp = Prefs.SyhiDepth;
      Prefs.SyhiDepth = G_MAXINT;
      refresh_syhi(GTK_EDITABLE(Editor), NULL);
      reconnect_sig(GTK_EDITABLE(Editor), NULL);
      Prefs.SyhiDepth = tmp;
    }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
}

/*
** This fcn return the number of the language of the current Editor
**
** Parameters :
**  void
**
** Return Values :
**  -1			Language is always unknown
** 0..MAX_LANG - 1	The language
**
** NOTE_WARNING: UEdit numbers languages from 1 to MAX_LANG
*/
gint			guess_lang(void)
{
  gchar			*ext_orig;
  gint			ext_len;
  gchar			*ext_up;
  gint			i;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  ext_orig = FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
		    Type);
#ifdef DEBUG_EXT_REC
  g_print(__FILE__": "__func__"(): Getting length of ext_orig...\n");
#endif
  ext_len = strlen(ext_orig);
  /* This macro FPROPS is awful :) */
#ifdef DEBUG_EXT_REC
  g_print(__FILE__": "__func__"(): Original extension = \"%s\", len = %i\n",
	  ext_orig, ext_len);
#endif
  ext_up = malloc(ext_len + 3); /* 1 space before, 1 after, and '\0' */
  strcpy(ext_up + 1, ext_orig);
  ext_up[0] = ' ';
  ext_up[ext_len + 1] = ' ';
  ext_up[ext_len + 2] = '\0';
  for (i = 1; i < ext_len + 1; i++)
    ext_up[i] = toupper(ext_up[i]);
  /* Now, ext_up contains the extension in upper case */
#ifdef DEBUG_EXT_REC
  g_print(__FILE__": "__func__"(): ' ' + EXTENSION + ' ': \"%s\"\n", ext_up);
#endif
  for (i = 0; i < MAX_LANG; i++)
    {
#ifdef DEBUG_EXT_REC
      g_print(__FILE__": "__func__"(): Ext of language /L%i = \"%s\"\n",
	      i + 1, Prefs.L[i].Extensions);
#endif
      if (Prefs.L[i].IsDefined && strstr(Prefs.L[i].Extensions, ext_up))
	{
	  free(ext_up);
#ifdef DEBUG_FCN
	  g_print(__FILE__": "__func__"(): End\n");
#endif
	  return i;
	}
    }
  free(ext_up);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return -1;
}


/*
** This fcn is used to read the UltraEdit wordfile to init different
** fields of the global variable Prefs
**
** Parameters :
**  wf_name		UEdit wordfile name
**
** Return Values :
**  0			All is right
**  -1			Can not open wordfile
**  -2			Parse error in a language section
*/
extern gint		read_uedit_wordfile(const gchar *wf_name)
{
  FILE			*fp;
  gint			filesize;
  gchar			*buffer;
  gint			i;
  gint			j;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  /* Initialisation of all variables */
  /* NOTE_TODO: Free all these variables by creating a new fcn called when
     exiting Beaver :) */
  for (i = 0; i < MAX_LANG; i++)
    {
      for (j = 0; j < MAX_COL; j++)
	{
	  Prefs.L[i].C[j].Keywords = NULL;
	  Prefs.L[i].C[j].Description[0] = '\0';
	}
      Prefs.L[i].IsDefined = FALSE;
      Prefs.L[i].Description[0] = '\0';
      Prefs.L[i].Extensions = NULL;
      Prefs.L[i].LineComment = NULL;
      Prefs.L[i].LineCommentAlt = NULL;
      Prefs.L[i].BlockCommentOn = NULL;
      Prefs.L[i].BlockCommentOff = NULL;
      Prefs.L[i].BlockCommentOnAlt = NULL;
      Prefs.L[i].BlockCommentOffAlt = NULL;
      Prefs.L[i].IsCaseSensitive = TRUE;
      Prefs.L[i].HaveString = TRUE;
      Prefs.L[i].StringChar0 = '\"';
      Prefs.L[i].StringChar1 = '\0';
      Prefs.L[i].EscapeChar = '\0';
      Prefs.L[i].Delimiters = NULL;
      Prefs.L[i].IndentString = NULL;
      Prefs.L[i].UnindentString = NULL;
      Prefs.L[i].MarkerChars = NULL;
      Prefs.L[i].FunctionString = NULL;
      Prefs.L[i].IsHTML = FALSE;
    }
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): Openning \"%s\"...\n", wf_name);
#endif
  if (!(fp = fopen(wf_name, "rt")))
    {
      gchar *wf_not_found;
      
      wf_not_found = g_strconcat("WARNING: Cannot find \"", wf_name,
				 "\"...", NULL);
      print_msg (wf_not_found);
      g_free (wf_not_found);
      return -1;
    }
  filesize = 0;
  while (getc(fp) != EOF)
    filesize++;
  if (filesize > 0)
    {
      buffer = g_malloc(filesize + 2);
      fseek(fp, 0, SEEK_SET);
      for (i = 0; i < filesize; i++)
	buffer[i] = getc(fp);
      buffer[filesize] = '\n';
      buffer[filesize + 1] = '\0';
    }
  else
    buffer = NULL;
  fclose(fp);
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): Wordfile size = %i\n", filesize);
#endif
  /* Here: main work */
  i = 0;
  while (i + 1 <= filesize)
    {
      if (!strncmp(buffer + i, "/L", 2))
	{
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Calling "
		  "parse_language_section()\n");
#endif
	  if (parse_language_section(buffer, filesize, &i))
	    {
	      g_print(__FILE__": "__func__"(): End, parse error in a "
		      "language section of \"%s\"\n", wf_name);
	      g_free(buffer);
	      return -2;
	    }
	}
      else
	i++;
    }
  g_free(buffer);
  for (i = 0; i < MAX_LANG; i++)
    {
      if (!Prefs.L[i].IsDefined)
	sprintf(Prefs.L[i].Description, "Language %i", i + 1);
      Prefs.L[i].Description[MAXLEN_LANG_DESCRIPTION] = 0;
      for (j = 0; j < MAX_COL; j++)
	{
	  if (!Prefs.L[i].C[j].Keywords)
	    sprintf(Prefs.L[i].C[j].Description, "Keywords Group %i", j + 1);
	  Prefs.L[i].C[j].Description[MAXLEN_COL_DESCRIPTION] = 0;
	}
    }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return 0;
}

/*
** Parse a language section in a UEdit wordfile
**
** Parameters:
**  buffer		Is a null-terminated string, representing the /Lx
**			section in a wordfile
**  size		Size of the buffer
**  start		buffer + *start points on "/Lx..."
**
** Return values:
**  0			All is right
**  -1			Parse error in a color section
*/
gint			parse_language_section(gchar *buffer,
					       gint size,
					       gint *start)
{
  gint			Lg;
  gint			i;
  gint			j;
  gchar			c;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  i = *start;
  i = i + 2;
  /* buffer[i] is the 1st digit of language number */
  j = i;
  while (isdigit(buffer[j]))
    j++;
  /* buffer[j] is the first char after the last digit of language number */
  c = buffer[j];
  buffer[j] = 0;
  Lg = atoi(buffer + i) - 1;
  buffer[j] = c;
  if (Lg + 1 > MAX_LANG)
    {
      g_print(__FILE__": "__func__"(): Warning, /L%i section detected "
	      "but only %i are allowed\n", Lg + 1, MAX_LANG);
      return -1;
    }
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): /L%i section detected\n", Lg + 1);
#endif
  Prefs.L[Lg].IsDefined = TRUE;
  i = j;
  if (buffer[i] == '\"')
    {
      /* There is a description of the language */
      j = ++i;
      while (buffer[j] != '\"')
	j++;
      /* buffer[i] is the 1st char of description and buffer[j] the 2nd '\"' */
      if ((j - i) > MAXLEN_LANG_DESCRIPTION)
	{
	  g_print(__FILE__": "__func__"(): Warning, description of "
		  "language %i is longer than %i\n", Lg + 1,
		  MAXLEN_LANG_DESCRIPTION);
	  return -1;
	}
      strncpy(Prefs.L[Lg].Description, buffer + i, j - i);
      Prefs.L[Lg].Description[j - i] = '\0';
      i = ++j;
    }
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): Description of language %i = "
	  "\"%s\"\n", Lg + 1, Prefs.L[Lg].Description);
#endif
  /* Here, buffer[i (= j)] is next non-identified char */
  /* The while loop below reads the line /Lx of a wordfile */
  while (buffer[i] != '\n')
    {
      /* LineComment */
      if (!strncmp(buffer + i, "Line Comment = ", 15))
	{
	  Prefs.L[Lg].LineComment = g_malloc(MAXLEN_LINE_COMMENT + 1);
	  i += 15;
	  j = -1;
	  while ((++j < MAXLEN_LINE_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].LineComment[j] = buffer[i + j];
	  if ((j == MAXLEN_LINE_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Line Comment of "
		      "language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_LINE_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].LineComment[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Line Comment = \"%s\"\n",
		  Prefs.L[Lg].LineComment);
#endif
	}
      /* LineCommentAlt */
      else if (!strncmp(buffer + i, "Line Comment Alt = ", 19))
	{
	  Prefs.L[Lg].LineCommentAlt = g_malloc(MAXLEN_LINE_COMMENT + 1);
	  i += 19;
	  j = -1;
	  while ((++j < MAXLEN_LINE_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].LineCommentAlt[j] = buffer[i + j];
	  if ((j == MAXLEN_LINE_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Line Comment Alt of "
		      "language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_LINE_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].LineCommentAlt[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Line Comment Alt = \"%s\"\n",
		  Prefs.L[Lg].LineCommentAlt);
#endif
	}
      /* LineComment Num */
      else if (!strncmp(buffer + i, "Line Comment Num = ", 19))
	{
	  gint		line_cmt_len;

	  Prefs.L[Lg].LineComment = g_malloc(MAXLEN_LINE_COMMENT + 1);
	  line_cmt_len = buffer[i += 19] - '0';
	  if (line_cmt_len > MAXLEN_LINE_COMMENT)
	    {
	      g_print(__FILE__": "__func__"(): Warning, Line Comment Num of "
		      "language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_LINE_COMMENT);
	      return -1;
	    }
	  i++;
	  /* Here buffer[i] is the 1st char of LineComment */
	  for (j = 0; j < line_cmt_len; j++)
	    Prefs.L[Lg].LineComment[j] = buffer[i + j];
	  Prefs.L[Lg].LineComment[j] = '\0';
	  i += line_cmt_len;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Line Comment Num = \"%s\"\n",
		  Prefs.L[Lg].LineComment);
#endif
	}
      /* BlockCommentOn */
      else if (!strncmp(buffer + i, "Block Comment On = ", 19))
	{
	  Prefs.L[Lg].BlockCommentOn = g_malloc(MAXLEN_BLOCK_COMMENT + 1);
	  i += 19;
	  j = -1;
	  while ((++j < MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].BlockCommentOn[j] = buffer[i + j];
	  if ((j == MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Block Comment On of "
		      "language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_BLOCK_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].BlockCommentOn[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Block Comment On = \"%s\"\n",
		  Prefs.L[Lg].BlockCommentOn);
#endif
	}
      /* BlockCommentOff */
      else if (!strncmp(buffer + i, "Block Comment Off = ", 20))
	{
	  Prefs.L[Lg].BlockCommentOff = g_malloc(MAXLEN_BLOCK_COMMENT + 1);
	  i += 20;
	  j = -1;
	  while ((++j < MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].BlockCommentOff[j] = buffer[i + j];
	  if ((j == MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Block Comment Off of "
		      "language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_BLOCK_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].BlockCommentOff[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Block Comment Off = \"%s\"\n",
		  Prefs.L[Lg].BlockCommentOff);
#endif
	}
      /* BlockCommentOnAlt */
      else if (!strncmp(buffer + i, "Block Comment On Alt = ", 23))
	{
	  Prefs.L[Lg].BlockCommentOnAlt = g_malloc(MAXLEN_BLOCK_COMMENT + 1);
	  i += 23;
	  j = -1;
	  while ((++j < MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].BlockCommentOnAlt[j] = buffer[i + j];
	  if ((j == MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Block Comment On Alt "
		      "of language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_BLOCK_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].BlockCommentOnAlt[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Block Comment On Alt = \"%s\"\n",
		  Prefs.L[Lg].BlockCommentOnAlt);
#endif
	}
      /* BlockCommentOffAlt */
      else if (!strncmp(buffer + i, "Block Comment Off Alt = ", 24))
	{
	  Prefs.L[Lg].BlockCommentOffAlt = g_malloc(MAXLEN_BLOCK_COMMENT + 1);
	  i += 24;
	  j = -1;
	  while ((++j < MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    Prefs.L[Lg].BlockCommentOffAlt[j] = buffer[i + j];
	  if ((j == MAXLEN_BLOCK_COMMENT) && (buffer[i + j] != ' '))
	    {
	      g_print(__FILE__": "__func__"(): Warning, Block Comment Off Alt "
		      "of language /L%i is longer than %i\n",
		      Lg + 1, MAXLEN_BLOCK_COMMENT);
	      return -1;
	    }
	  Prefs.L[Lg].BlockCommentOffAlt[j] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Block Comment Off Alt = \"%s\"\n",
		  Prefs.L[Lg].BlockCommentOffAlt);
#endif
	}
      /* StringChars */
      else if (!strncmp(buffer + i, "String Chars = ", 15))
	{
	  i += 15;
	  if (buffer[i] != ' ')
	    Prefs.L[Lg].StringChar0 = buffer[i];
	  i++;
	  if (buffer[i] != ' ')
	    Prefs.L[Lg].StringChar1 = buffer[i];
	  i++;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): String Chars = \'%c\' & \'%c\'\n",
		  Prefs.L[Lg].StringChar0, Prefs.L[Lg].StringChar1);
#endif
	}
      /* EscapeChar */
      else if (!strncmp(buffer + i, "Escape Char = ", 14))
	{
	  i += 14;
	  if (buffer[i] != ' ')
	    Prefs.L[Lg].EscapeChar = buffer[i];
	  i++;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Escape Char = \'%c\'\n",
		  Prefs.L[Lg].EscapeChar);
#endif
	}
      /* FileExtensions */
      /* NOTE_WARNING: This sections is supposed to end the line /Lx */
      else if (!strncmp(buffer + i, "File Extensions = ", 18))
	{
	  i += 18;
	  j = 0;
	  while (buffer[i + j] != '\n')
	    j++;
	  Prefs.L[Lg].Extensions = g_malloc(1 + j + 1 + 1);
	  Prefs.L[Lg].Extensions[0] = ' ';
	  j = 0;
	  while (buffer[i + j] != '\n')
	      Prefs.L[Lg].Extensions[1 + j] = buffer[i + j++];
	  Prefs.L[Lg].Extensions[j] = ' ';
	  Prefs.L[Lg].Extensions[j + 1] = '\0';
	  i += j;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): File Extensions = \"%s\"\n",
		  Prefs.L[Lg].Extensions);
#endif
	}
      /* Nocase */
      else if (!strncmp(buffer + i, "Nocase ", 7))
	{
	  i += 7;
	  Prefs.L[Lg].IsCaseSensitive = FALSE;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Nocase\n");
#endif
	}
      /* Noquote */
      else if (!strncmp(buffer + i, "Noquote ", 8))
	{
	  i += 8;
	  Prefs.L[Lg].HaveString = FALSE;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Noquote\n");
#endif
	}
      /* HTML */
      else if (!strncmp(buffer + i, "HTML_LANG ", 10))
	{
	  i += 10;
	  Prefs.L[Lg].IsHTML = TRUE;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): HTML_LANG\n");
#endif
	}
      /* FORTRAN_LANG */
      /* NOTE_TODO: Implement it: 'C', 'c' or '*' in the 1st column is a
	 line comment*/
      else if (!strncmp(buffer + i, "FORTRAN_LANG ", 13))
	{
	  i += 13;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): FORTRAN_LANG\n");
#endif
	}
      /* LATEX_LANG */
      /* NOTE_REMARK: LATEX_LANG is ignored, Beaver does not need it ! */
      else if (!strncmp(buffer + i, "LATEX_LANG ", 11))
	{
	  i += 11;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): LATEX_LANG\n");
#endif
	}
      else if ((buffer[i] == ' ') || (buffer[i] == '\t'))
	i++;
      /* Nothing */
      else
	{
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Unreconized string in /Lx line, "
		  "beginning = \"%c%c%c%c...\"\n",
		  buffer[i], buffer[i + 1], buffer[i + 2], buffer[i + 3]);
#endif
	  i++;
	}
    } /* while (buffer[i] != '\n') */
  i++;
  while ((i + 1 < size) && strncmp(buffer + i, "/C", 2))
    {
      if (!strncmp(buffer + i, "/Delimiters = ", 14))
	{
	  /* Delimiters */
	  j = (i += 14);
	  while (buffer[j] != '\n')
	    j++;
	  j++;
	  /* NOTE_WARNING: Yes, the '\n' needs to be into the Delimiters */
	  Prefs.L[Lg].Delimiters = g_malloc(j - i + 1);
	  strncpy(Prefs.L[Lg].Delimiters, buffer + i, j - i);
	  Prefs.L[Lg].Delimiters[j - i] = '\0';
	  i = j;
	  for (j = 0; j < 256; j++)
	    Prefs.L[Lg].IsADelimiter[j] =
	      strchr(Prefs.L[Lg].Delimiters, j) ? 1 : 0;
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Delimiters = \"%s\"\n",
		  Prefs.L[Lg].Delimiters);
#endif
	}
      else if (!strncmp(buffer + i, "/Indent Strings =", 17))
	{
	  /* Indent Strings */
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Indent String = ");
#endif
	  i += 17;
	  while ((buffer[i] != '\n') && (buffer[i] != '\"'))
	    i++;
	  if (buffer[i] == '\"')
	    {
	      j = ++i;
	      while ((buffer[j] != '\n') && (buffer[j] != '\"'))
		j++;
	      /* Here, if buffer[j] == '\n', it is considered as a '\"' */
	      Prefs.L[Lg].IndentString = g_malloc(j - i + 1);
	      strncpy(Prefs.L[Lg].IndentString, buffer + i, j - i);
	      Prefs.L[Lg].IndentString[j - i] = '\0';
	      i = j;
	      /* NOTE_TODO : Support multiple indent string... */
	      while (buffer[i] != '\n')
		i++;
	      i++;
#ifdef DEBUG_WORDFILE
	      g_print("\"%s\"\n", Prefs.L[Lg].IndentString);
#endif
	    }
	  else
	    i++;
	}
      else if (!strncmp(buffer + i, "/Unindent Strings =", 19))
	{
	  /* Unindent Strings */
	  i += 19;
	  while ((buffer[i] != '\n') && (buffer[i] != '\"'))
	    i++;
	  if (buffer[i] == '\"')
	    {
	      j = ++i;
	      while ((buffer[j] != '\n') && (buffer[j] != '\"'))
		j++;
	      /* Here, if buffer[j] == '\n', it is considered as a '\"' */
	      Prefs.L[Lg].UnindentString = g_malloc(j - i + 1);
	      strncpy(Prefs.L[Lg].UnindentString, buffer + i, j - i);
	      Prefs.L[Lg].UnindentString[j - i] = '\0';
	      i = j;
	      /* NOTE_TODO : Support multiple unindent string... */
	      while (buffer[i] != '\n')
		i++;
	      i++;
#ifdef DEBUG_WORDFILE
	      g_print(__FILE__": "__func__"(): Unindent String = \"%s\"\n",
		      Prefs.L[Lg].UnindentString);
#endif
	    }
	  else
	    i++;
	}
      else if (!strncmp(buffer + i, "/Marker Characters =", 20))
	{
	  /* Marker Characters */
#ifdef DEBUG_WORDFILE
	      g_print(__FILE__": "__func__"(): Marker Chars = ");
#endif
	  i += 20;
	  while ((buffer[i] != '\n') && (buffer[i] != '\"'))
	    i++;
	  if (buffer[i] == '\"')
	    {
	      j = ++i;
	      while ((buffer[j] != '\n') && (buffer[j] != '\"'))
		j++;
	      /* Here, if buffer[j] == '\n', it is considered as a '\"' */
	      Prefs.L[Lg].MarkerChars = g_malloc(j - i + 1);
	      strncpy(Prefs.L[Lg].MarkerChars, buffer + i, j - i);
	      Prefs.L[Lg].MarkerChars[j - i] = '\0';
	      i = j;
	      /* NOTE_TODO : Support marker chars... */
	      while (buffer[i] != '\n')
		i++;
	      i++;
#ifdef DEBUG_WORDFILE
	      g_print("\"%s\"\n", Prefs.L[Lg].MarkerChars);
#endif
	    }
	  else
	    i++;
	}
      else if (!strncmp(buffer + i, "/Function String =", 18))
	{
	  /* Function String */
#ifdef DEBUG_WORDFILE
	      g_print(__FILE__": "__func__"(): Function String = ");
#endif
	  i += 18;
	  while ((buffer[i] != '\n') && (buffer[i] != '\"'))
	    i++;
	  if (buffer[i] == '\"')
	    {
	      j = ++i;
	      while ((buffer[j] != '\n') && (buffer[j] != '\"'))
		j++;
	      /* Here, if buffer[j] == '\n', it is considered as a '\"' */
	      Prefs.L[Lg].FunctionString = g_malloc(j - i + 1);
	      strncpy(Prefs.L[Lg].FunctionString, buffer + i, j - i);
	      Prefs.L[Lg].FunctionString[j - i] = '\0';
	      i = j;
	      /* NOTE_TODO : Support function string */
	      while (buffer[i] != '\n')
		i++;
	      i++;
#ifdef DEBUG_WORDFILE
	      g_print("\"%s\"\n", Prefs.L[Lg].FunctionString);
#endif
	    }
	  else
	    i++;
	}
      else if ((buffer[i] == ' ') || (buffer[i] == '\t'))
	i++;
      else
	{
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): Unregonized string between /Lx "
		  "and /Cy: \"%c%c%c%c...\"\n",
		  buffer[i], buffer[i + 1], buffer[i + 2], buffer[i + 3]);
#endif
	  i++;
	}
    } /* while (strncmp(buffer + i, "/C", 2)) */
  while ((i + 1 <= size) && strncmp(buffer + i, "/L", 2))
    {
      if (!strncmp(buffer + i, "/C", 2))
	{
	  if (parse_color_section(buffer, size, &i, Lg))
	    {
	      g_print(__FILE__": "__func__"(): End, parse error in a color "
		      "section of /L%i\n", Lg + 1);
	      return -1;
	    }
	}
      else
	i++;
    }
  *start = i;
  /*
  ** Now all Prefs.L[Lg].C[_var_].Keywords have been initialized, we can
  ** construct the wtree...
  **
  ** NOTE_TODO: free this memory allocated when exit of beaver
  */
#ifdef DEBUG_WTREE
  g_print(__FILE__": "__func__"(): Constructing wtree (Lg=%d)...\n", Lg);
#endif
  if ((Prefs.L[Lg].wtree_root = g_malloc(sizeof (*(Prefs.L[Lg].wtree_root)))))
    {
      t_wtree		*walk;
      guchar		uc;
      int		size_lg;

      size_lg = sizeof (*(Prefs.L[Lg].wtree_root));
      memset(Prefs.L[Lg].wtree_root, 0, sizeof (*(Prefs.L[Lg].wtree_root)));
#ifdef DEBUG_WTREE
      g_print(__FILE__": "__func__"(): Total memory size of wtree (Lg=%d): "
	      "%d KB\n", Lg, size_lg / 1024);
#endif
      for (i = 0; i < MAX_COL; i++)
	if (Prefs.L[Lg].C[i].Keywords)
	  {
	    j = 1;
	    while (Prefs.L[Lg].C[i].Keywords[j])
	      {
		walk = Prefs.L[Lg].wtree_root;
		while ((uc = Prefs.L[Lg].C[i].Keywords[j]) != ' ')
		  {
		    if (uc >= 128)
		      {
			j++;
			continue;
		      }
		    /*
		    ** NOTE_REMARK: Here, if Prefs.L[Lg].IsCaseSensitive != 0,
		    **  we store its keywords in lower-case
		    */
		    if (!Prefs.L[Lg].IsCaseSensitive)
		      uc = (uc >= 'A' && uc <= 'Z') ? (uc | 0x20) : (uc);
		    if (!walk->letter[uc])
		      {
			if ((walk->letter[uc] =
			     g_malloc(sizeof (*walk->letter[uc]))))
			  {
			    /*
			    ** NOTE_REMARK: And (if Prefs.L[Lg].
			    **  IsCaseSensitive is always != 0) we copy the
			    **  pointer to next node at the index of
			    **  the array 'letter' corresponding to the
			    **  uppercase letter
			    */
			    if (!Prefs.L[Lg].IsCaseSensitive &&
				(uc >= 'a' && uc <= 'z'))
			      walk->letter[uc & ~0x20] = walk->letter[uc];
			    size_lg += sizeof (*(walk->letter[uc]));
			    memset(walk->letter[uc], 0,
				   sizeof (*(walk->letter[uc])));
#ifdef DEBUG_WTREE
			    g_print(__FILE__": "__func__"(): Total memory size"
			    " of wtree (Lg=%d): %d KB\n", Lg, size_lg / 1024);
#endif
			  }
			else
			  {
#ifdef DEBUG_WTREE
			    g_print(__FILE__": "__func__"(): malloc failed\n");
#endif
			    return (-1);
			  }
		      }
		    walk = walk->letter[uc];
		    j++;
		  } /* while (uc != ' ') */
		j++;
		walk->color_plus_1 = i + 1;
	      } /* while (uc) */
	  } /* if (Prefs....) */
#ifdef DEBUG_WTREE
      g_print(__FILE__": "__func__"(): wtree (Lg=%d) is made (%d KB)\n",
	      Lg, size_lg / 1024);
#endif
    }
  else
    {
#ifdef DEBUG_WTREE
      g_print(__FILE__": "__func__"(): g_malloc() of wtree_root failed\n");
#endif
    }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return 0;
}

/*
** Parse of color section in a UEdit wordfile
**
** Parameters:
**  buffer		Is a null-terminated string, representing the /Cx
**			section in a wordfile
**  size		Size of the buffer
**  start		buffer + *start points on "/Cx..."
**
** Return values:
**  0			No errors
**  -1			Parse error in a line
*/
gint			parse_color_section(gchar *buffer,
					    gint size,
					    gint *start,
					    gint Lg)
{
  gint			col;
  gint			i;
  gint			j;
  gchar			c;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  i = *start;
  i = i + 2;
  /* buffer[i] is the 1st digit of color number */
  j = i;
  while (isdigit(buffer[j]))
    j++;
  /* buffer[j] is the first char after the last digit of language number */
  c = buffer[j];
  buffer[j] = 0;
  col = atoi(buffer + i) - 1;
  buffer[j] = c;
  if (col + 1 > MAX_COL)
    {
      g_print(__FILE__": "__func__"(): Warning, color %i section detected "
	      "but only %i are allowed\n", col + 1, MAX_COL);
      return -1;
    }
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): Color %i section detected\n", col + 1);
#endif
  i = j;
  if (buffer[i] == '\"' || buffer[i] == ' ')
    {
      /* There is a description of the color */
      j = ++i;
      while (buffer[j] != '\"' && buffer[j] != '\n')
	j++;
      /* buffer[i] is the 1st char of description and buffer[j] the 2nd '\"' */
      if ((j - i) > MAXLEN_COL_DESCRIPTION)
	{
	  g_print(__FILE__": "__func__"(): Warning, description of color %i "
		  "of language %i is longer than %i\n", col + 1, Lg + 1,
		  MAXLEN_COL_DESCRIPTION);
	  return -1;
	}
      strncpy(Prefs.L[Lg].C[col].Description, buffer + i, j - i);
      Prefs.L[Lg].C[col].Description[j - i] = '\0';
      i = j;
    }
#ifdef DEBUG_WORDFILE
  g_print(__FILE__": "__func__"(): Description of color %i = "
	  "\"%s\"\n", col + 1, Prefs.L[Lg].C[col].Description);
#endif
  while (buffer[i] != '\n')
    i++;
  j = ++i;
  /* Here, buffer[i] is the first char of keywords area */
  while ((i + 1 <= size) &&
	 strncmp(buffer + i, "/C", 2) &&
	 strncmp(buffer + i, "/L", 2))
    {
      if ((buffer[i] == '/') && (buffer[i + 1] != '/'))
	{
	  /* NOTE_TODO: Take into account this case (execute the 'command') */
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): In /L%i /C%i, command ignored\n",
		  Lg + 1, col + 1);
#endif
	  while (buffer[i] != '\n')
	    i++;
	  i++;
	}
      else  if (!strncmp(buffer + i, "**", 2))
	{
	  /* NOTE_TODO: Take into account this case (colorize all words
	     beginning with following prefixes) */
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): In /L%i, '**' ignored\n",
		  Lg + 1);
#endif
	  while (buffer[i] != '\n')
	    i++;
	  i++;
	}
      else
	{
	  if ((buffer[i] == '/') && (buffer[i + 1] == '/'))
	    {
#ifdef DEBUG_WORDFILE
	      g_print(__FILE__": "__func__"(): In /L%i, '//' detected\n",
		      Lg + 1);
#endif
	      i += 2;
	    }
	  if (parse_line_of_keywords(buffer, size, &i, Lg, col))
	    {
	      g_print(__FILE__": "__func__"(): End, parse error in a line of "
		      "keywords of /L%i /C%i\n", Lg + 1, col + 1);
	      return -1;
	    }
	}
    }
  *start = i;
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return 0;
}

/*
** Parse 1 line of keywords in a UEdit wordfile
**
** Parameters:
**  buffer		Is a null-terminated string, representing the /Cx
**			section in a wordfile
**  size		Size of the buffer
**  start		buffer + *start points on "/Cx..."
**
** Return values:
**  0			No errors
**  -1			Error while parsing line
*/
gint			parse_line_of_keywords(gchar *buffer,
					       gint size,
					       gint *start,
					       gint Lg,
					       gint col)
{
  gint			i;
  gint			j;
  gchar			c;
  gint			strlen_k;
  gint			strlen_b;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  (void)size;
  i = *start;
  while ((buffer[i] != '\r') && (buffer[i] != '\n'))
    {
      if ((buffer[i] == ' ') || (buffer[i] == '\t'))
	  i++;
      else
	{
	  /* buffer[i] is the first char of the keyword */
	  j = i;
	  while ((buffer[j] != ' ') && (buffer[j] != '\t') &&
		 (buffer[j] != '\r') && (buffer[j] != '\n'))
	    j++;
	  /* buffer[j] is the first char after the last one of the keyword */
	  c = buffer[j];
	  buffer[j] = '\0';
#ifdef DEBUG_WORDFILE
	  g_print(__FILE__": "__func__"(): In /L%i /C%i, keyword \"%s\" "
		  "detected\n", Lg + 1, col + 1, buffer + i);
#endif
	  if (Prefs.L[Lg].C[col].Keywords == NULL)
	    {
#ifdef DEBUG_WORDFILE
	      g_print(__FILE__": "__func__"(): Ah ! Keywords == NULL\n");
#endif
	      if (!(Prefs.L[Lg].C[col].Keywords = g_malloc(2)))
		{
		  g_print(__FILE__": "__func__"(): End, g_malloc() failed\n");
		  return -1;
		}
	      Prefs.L[Lg].C[col].Keywords[0] = ' ';
	      Prefs.L[Lg].C[col].Keywords[1] = '\0';
	    }
	  strlen_k = strlen(Prefs.L[Lg].C[col].Keywords);
	  strlen_b = strlen(buffer + i);
	  Prefs.L[Lg].C[col].Keywords =
	    g_realloc(Prefs.L[Lg].C[col].Keywords, strlen_k + strlen_b + 2);
	  if (Prefs.L[Lg].C[col].Keywords == NULL)
	    {
	      g_print(__FILE__": "__func__"(): End, g_realloc() failed\n");
	      return -1;
	    }
	  strcat(Prefs.L[Lg].C[col].Keywords, buffer + i);
	  strcat(Prefs.L[Lg].C[col].Keywords, " ");
	  buffer[j] = c;
	  i = j;
	} /* !((buffer[i] == ' ') || (buffer[i] == '\t')) */
    }
  if (buffer[i] == '\r')
    i++;
  *start = i + 1;
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return 0;
}

/*
** Callback fcn for insert_text signal
**
** Parameters :
**  See a gtk reference manual
**
** Return values :
**  TRUE		Always
**  FALSE		Never
**
** NOTE_REMARK : the signal is generated only by the keyboard (not by
**  gtk_text_insert() fcn but by gtk_editable_insert_text) and before the
**  insertion of text in the widget
*/
gboolean		text_has_been_inserted(GtkEditable *Editor,
					       const gchar *Text,
					       gint length,
					       gint *position,
					       gpointer data)
{
  t_action		*action;
  int			Lg;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  /* NOTE_WARNING: Text is not nul-terminated !!!!! Probably a GtkEditable
     Bug !! */
  (void)data;
  Lg = FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	      WidgetInfo.Lg);
  /* NOTE_WARNING : Without affecting *position in the next line of code,
     there is a big bug :
     Selecting text with mouse, and then pressing a key would delete selected
     text, and then would insert the key pressed at the end of text...
     Unbelievable ! Is it a GTK bug ? Probably, because there is not
     this problem, when something is pasted in replacement of hitting a key. */
  *position = refresh_from = gtk_editable_get_position(Editor);
  /* refresh_from is the index where Text will be inserted.
     Now checking for auto-correction or auto-indentation */
  if (length == 1)
    {
      /* If AutoCorrection and if the char inserted is a Delimiter,
	 word_to_correct should be the index of the word to auto-correct.
	 But here, it points only on the first char after the word to be
	 corrected, final computing of its right value will be done in
	 refresh_syhi() */
      if (Prefs.AutoCorrection &&
	  Prefs.L[Lg].Delimiters && strchr(Prefs.L[Lg].Delimiters, *Text))
	word_to_correct = refresh_from;
      if (*Text == '\n')
	enter_pressed = TRUE;
    }
  action = malloc(sizeof (t_action));
  action->type = insert;
  action->start = refresh_from;
  action->end = refresh_from + length;
  action->text = malloc(length + 1);
  strncpy(action->text, Text, length);
  action->text[length] = 0;
  record_action(action);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return TRUE;
}

/*
** Callback fcn for delete_text signal
**
** Parameters :
**  See a gtk reference manual
**
** Return values :
**  TRUE		Always
**  FALSE		Never
**
** NOTE_REMARK : the signal is generated only by the keyboard (not by
** gtk_text_*ward_delete() fcn. NOTE_WARNING : but by gtk_editable_delete_text)
** and before the deletion of text in the widget
*/
gboolean		text_has_been_deleted(GtkEditable *Editor,
					      gint start,
					      gint end,
					      gpointer data)
{
  t_action		*action;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  (void)Editor;
  (void)end;
  (void)data;
  refresh_from = start;
  action = malloc(sizeof *action);
  action->type = delete;
  action->start = start;
  action->end = end;
  action->text = gtk_editable_get_chars(Editor, start, end);
  record_action(action);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return TRUE;
}

/*
** Callback fcn for changed signal. It refresh syntax highlight from
** char in editor indexed by refresh_from
**
** Parameters :
**  See a gtk reference manual
**
** Return values :
**  TRUE		Always
**  FALSE		Never
*/
gboolean		refresh_syhi(GtkEditable *Editor, gpointer data)
{
  gchar			*buffer;
  gint			buflen;
  t_char_state		refresh_from_state;
  gint			i, j;
  guint			CursorPosition;
  gint			Lg;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  (void)data;
  Lg = FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	      WidgetInfo.Lg);
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Language of file modified: /L%i \"%s\"\n",
	  Lg + 1, Prefs.L[Lg].Description);
#endif
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectIns));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectDel));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectChg));
  gtk_signal_disconnect(GTK_OBJECT(Editor),
			FPROPS(gtk_notebook_get_current_page
			       (GTK_NOTEBOOK(MainNotebook)),
			       WidgetInfo.SigConnectChgAft));
  gtk_text_freeze(GTK_TEXT(Editor));
  /* Here is beginning the real work... */
  buffer = gtk_editable_get_chars(GTK_EDITABLE(Editor), 0, -1);
  buflen = strlen(buffer);
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Length of buffer = %i\n", buflen);
#endif
  /* Finishing computing the good value for refresh_from */
  if (enter_pressed)
    refresh_from--;
  refresh_from = (refresh_from > 0) ? refresh_from - 1 : 0;
  while ((refresh_from > 0) && (buffer[refresh_from] != '\n'))
    refresh_from--;
  if (buffer[refresh_from] == '\n')
    refresh_from++;
  /* Here, refresh_from points to the first character of the line
     needed to be syhi-updated (where chars have been inserted or deleted).
     Or if RET has been pressed, refresh_from points on the new line */
  /* Now, Final computing of the right value for word_to_correct
     (Beginning in text_has_been_inserted())
     NOTE_REMINDER: word_to_correct has taken the value of cursor position
     before insertion of the delimiter */
#ifdef DEBUG_CORRECT
  g_print(__FILE__": "__func__"(): "
	  "Before: word_to_correct = %i [%c][%c][%c]\n",
	  word_to_correct,
	  ((buffer[word_to_correct + 0] >= ' ') &&
	  (buffer[word_to_correct + 0] <= '~')) ?
	  buffer[word_to_correct + 0] : 0x7f,
	  ((buffer[word_to_correct + 1] >= ' ') &&
	  (buffer[word_to_correct + 1] <= '~')) ?
	  buffer[word_to_correct + 1] : 0x7f,
	  ((buffer[word_to_correct + 2] >= ' ') &&
	  (buffer[word_to_correct + 2] <= '~')) ?
	  buffer[word_to_correct + 2] : 0x7f);
#endif
  word_to_correct--; /* Points on the last char of supposed word to correct */
  if ((word_to_correct > -1) &&
      !strchr(Prefs.L[Lg].Delimiters, buffer[word_to_correct]))
    {
      /* If the char just before the Delimiters inserted is not
	 another Delimiter (so is the end of the word to correct) */
      while ((word_to_correct >= 0) &&
	     !strchr(Prefs.L[Lg].Delimiters, buffer[word_to_correct]))
	word_to_correct--;
      word_to_correct++;
      /* Here, word_to_auto_correct points on the 1st char of
	 the word to auto-correct */
    }
  else
    word_to_correct = -1;
#ifdef DEBUG_CORRECT
  g_print(__FILE__": "__func__"(): "
	  "After : word_to_correct = %i [%c][%c][%c]\n",
	  word_to_correct,
	  ((buffer[word_to_correct + 0] >= ' ') &&
	  (buffer[word_to_correct + 0] <= '~')) ?
	  buffer[word_to_correct + 0] : 0x7f,
	  ((buffer[word_to_correct + 1] >= ' ') &&
	  (buffer[word_to_correct + 1] <= '~')) ?
	  buffer[word_to_correct + 1] : 0x7f,
	  ((buffer[word_to_correct + 2] >= ' ') &&
	  (buffer[word_to_correct + 2] <= '~')) ?
	  buffer[word_to_correct + 2] : 0x7f);
#endif
  i = 0;
  refresh_from_state = Normal;
  /* Now, computing refresh_from_state */
  while (i < refresh_from)
    {
      /*** Test for LineComment */
      if ((Prefs.L[Lg].LineComment != NULL) &&
	  !strncmp(buffer + i, Prefs.L[Lg].LineComment,
		   strlen(Prefs.L[Lg].LineComment)))
	{
	  /* NOTE_WARNING: I suppose there is no '\n' in LineComment */
	  while ((i < refresh_from) && (buffer[i] != '\n'))
	    i++;
	  i++;
	  refresh_from_state = Normal;
	}
      /*** Test for LineCommentAlt */
      else if ((Prefs.L[Lg].LineCommentAlt != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].LineCommentAlt,
			strlen(Prefs.L[Lg].LineCommentAlt)))
	{
	  /* NOTE_WARNING: I suppose there is no '\n' in LineComment */
	  while ((i < refresh_from) && (buffer[i] != '\n'))
	    i++;
	  i++;
	  refresh_from_state = Normal;
	}
      /*** Test for BlockCommentOn */
      else if ((Prefs.L[Lg].BlockCommentOn != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].BlockCommentOn,
			strlen(Prefs.L[Lg].BlockCommentOn)))
	{
	  if (Prefs.L[Lg].BlockCommentOff == NULL)
	    {
	      /* NOTE_WARNING: I suppose that BlockCommentOn does not contain
		 '\n' */
	      while ((i < refresh_from) && (buffer[i] != '\n'))
		i++;
	      i++;
	      refresh_from_state = Normal;
	    }
	  else if ((gint)NULL ==
		   (j = (gint)strstr(buffer + i +
				     strlen(Prefs.L[Lg].BlockCommentOn),
				     Prefs.L[Lg].BlockCommentOff)))
	    {
	      i = refresh_from;
	      refresh_from_state = Comment;
	    }
	  else
	    {
	      i = j + strlen(Prefs.L[Lg].BlockCommentOff) - (gint)buffer;
	      if (i <= refresh_from)
		refresh_from_state = Normal;
	      else
		refresh_from_state = Comment;
	    }
	}
      /*** Test for BlockCommentOnAlt */
      else if ((Prefs.L[Lg].BlockCommentOnAlt != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].BlockCommentOnAlt,
			strlen(Prefs.L[Lg].BlockCommentOnAlt)))
	{
	  if (Prefs.L[Lg].BlockCommentOffAlt == NULL)
	    {
	      /* NOTE_WARNING: I suppose that BlockCommentOnAlt does not
		 contain '\n' */
	      while ((i < refresh_from) && (buffer[i] != '\n'))
		i++;
	      i++;
	      refresh_from_state = Normal;
	    }
	  else if ((gint)NULL == 
		   (j = (gint)strstr(buffer + i +
				     strlen(Prefs.L[Lg].BlockCommentOnAlt),
				     Prefs.L[Lg].BlockCommentOffAlt)))
	    {
	      i = refresh_from;
	      refresh_from_state = CommentAlt;
	    }
	  else
	    {
	      i = j + strlen(Prefs.L[Lg].BlockCommentOffAlt) - (gint)buffer;
	      if (i <= refresh_from)
		refresh_from_state = Normal;
	      else
		refresh_from_state = CommentAlt;
	    }
	}
      /*** Test for StringChar0 */
      else if (Prefs.L[Lg].HaveString &&
	       (Prefs.L[Lg].StringChar0 != '\0') &&
	       (buffer[i] == Prefs.L[Lg].StringChar0))
	{
	  i++;
	  while ((i < refresh_from) && (buffer[i] != Prefs.L[Lg].StringChar0))
	    {
	      if (buffer[i] == Prefs.L[Lg].EscapeChar)
		i += 2;
	      else
		i++;
	    }
	  /* Here, i < refresh_from and points to the 2nd (closing) StringChar0
	     or i >= refresh_from: there is no 2nd (closing) StringChar0 until
	     i or refresh_from points to the 2nd (closing) StringChar0 */
	  if (i < refresh_from)
	    {
	      i++;
	      refresh_from_state = Normal;
	    }
	  else
	    {
	      i = refresh_from;
	      refresh_from_state = String0;
	    }
	}
      /*** Test for StringChar1 */
      else if (Prefs.L[Lg].HaveString &&
	       (Prefs.L[Lg].StringChar1 != '\0') &&
	       (buffer[i] == Prefs.L[Lg].StringChar1))
	{
	  i++;
	  while ((i < refresh_from) && (buffer[i] != Prefs.L[Lg].StringChar1))
	    {
	      if (buffer[i] == Prefs.L[Lg].EscapeChar)
		i += 2;
	      else
		i++;
	    }
	  /* Here, i < refresh_from and points to the 2nd (closing) StringChar1
	     or i >= refresh_from: there is no 2nd (closing) StringChar1 until
	     i or refresh_from points to the 2nd (closing) StringChar1 */
	  if (i < refresh_from)
	    {
	      i++;
	      refresh_from_state = Normal;
	    }
	  else
	    {
	      i = refresh_from;
	      refresh_from_state = String1;
	    }
	}
      /*** All tests failed */
      else
	{
	  i++;
	}
    } //while (i < refresh_from)
  /* Now, refresh_from_state is the state of the character pointed by buffer +
     refresh_from */
  /* NOTE_WARNING : this do not work !!!: CursorPosition =
     gtk_text_get_point(GTK_TEXT(Editor)); */
  CursorPosition = gtk_editable_get_position(GTK_EDITABLE(Editor));
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): CursorPosition = %i, deleting last chars\n",
	  CursorPosition);
#endif
  i = refresh_from;
  gtk_editable_delete_text(GTK_EDITABLE(Editor), refresh_from, -1);
  /* Now, we have to insert the text from buffer + i
     (at state refresh_from_state) to first char which state is Normal */
  /*** refresh_from is in Comment */
  if (refresh_from_state == Comment)
    {
      if (Prefs.L[Lg].BlockCommentOff == NULL)
	{
	  if ((j = (gint)strstr(buffer + i, "\n")) != (gint)NULL)
	    {
	      j = j + 1 - (gint)buffer;
	      gtk_text_insert(GTK_TEXT(Editor), NULL,
			      &(Prefs.Colors.L[Lg].Comment), NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor), NULL,
			      &(Prefs.Colors.L[Lg].Comment), NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
      else
	{
	  if ((j = (gint)strstr(buffer + i, Prefs.L[Lg].BlockCommentOff)) !=
	      (gint)NULL)
	    {
	      j = j + strlen(Prefs.L[Lg].BlockCommentOff) - (gint)buffer;
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL, &(Prefs.Colors.L[Lg].Comment),
			      NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL, &(Prefs.Colors.L[Lg].Comment),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
    }
  /*** refresh_from is in CommentAlt */
  else if (refresh_from_state == CommentAlt)
    {
      if (Prefs.L[Lg].BlockCommentOffAlt == NULL)
	{
	  if ((j = (gint)strstr(buffer + i, "\n")) != (gint)NULL)
	    {
	      j = j + 1 - (gint)buffer;
	      gtk_text_insert(GTK_TEXT(Editor), NULL,
			      &(Prefs.Colors.L[Lg].CommentAlt), NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL, &(Prefs.Colors.L[Lg].CommentAlt),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
      else
	{
	  if ((j = (gint)strstr(buffer + i, Prefs.L[Lg].BlockCommentOffAlt)) !=
	      (gint)NULL)
	    {
	      j = j + strlen(Prefs.L[Lg].BlockCommentOffAlt) - (gint)buffer;
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL, &(Prefs.Colors.L[Lg].CommentAlt),
			      NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL, &(Prefs.Colors.L[Lg].CommentAlt),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
    }
  /*** refresh_from is in String0 */
  else if (Prefs.L[Lg].HaveString && (refresh_from_state == String0))
    {
      j = i;
      while ((j < buflen) && (buffer[j] != Prefs.L[Lg].StringChar0))
	{
	  if (buffer[j] == Prefs.L[Lg].EscapeChar)
	    j += 2;
	  else
	    j++;
	}
      if (j >= buflen)
	{
	  gtk_text_insert(GTK_TEXT(Editor),
			  NULL, &(Prefs.Colors.L[Lg].String0), NULL,
			  buffer + i, -1);
	  i = buflen;
	}
      else
	{
	  gtk_text_insert(GTK_TEXT(Editor),
			  NULL, &(Prefs.Colors.L[Lg].String0), NULL,
			  buffer + i, j + 1 - i);
	  i = j + 1;
	}
    }
  /*** refresh_from is in String1 */
  else if (Prefs.L[Lg].HaveString && (refresh_from_state == String1))
    {
      j = i;
      while ((j < buflen) && (buffer[j] != Prefs.L[Lg].StringChar1))
	{
	  if (buffer[j] == Prefs.L[Lg].EscapeChar)
	    j += 2;
	  else
	    j++;
	}
      if (j >= buflen)
	{
	  gtk_text_insert(GTK_TEXT(Editor),
			  NULL, &(Prefs.Colors.L[Lg].String1), NULL,
			  buffer + i, -1);
	  i = buflen;
	}
      else
	{
	  gtk_text_insert(GTK_TEXT(Editor),
			  NULL, &(Prefs.Colors.L[Lg].String1), NULL,
			  buffer + i, j + 1 - i);
	  i = j + 1;
	}
    }
  /* Now, buffer + i points on '\0' (all have been inserted), or it points on
     a char, which state is "Normal"
     We have to continue to fill in the widget */
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Text between refresh_from and 1st state "
	  "Normal has been inserted\n");
#endif
  /* Old: while (i < buflen) */
  while (i < buflen && i < CursorPosition + Prefs.SyhiDepth)
    {
      /*** Detection of BlockCommentOn */
      if ((Prefs.L[Lg].BlockCommentOn != NULL) &&
	  !strncmp(buffer + i, Prefs.L[Lg].BlockCommentOn,
		   strlen(Prefs.L[Lg].BlockCommentOn)))
	{
	  if (Prefs.L[Lg].BlockCommentOff == NULL)
	    {
	      if ((j = (gint)strstr(buffer + i, "\n")) != (gint)NULL)
		{
		  j = j + 1 - (gint)buffer;
		  gtk_text_insert(GTK_TEXT(Editor), NULL,
				  &(Prefs.Colors.L[Lg].Comment), NULL,
				  buffer + i, j - i);
		  i = j;
		}
	      else
		{
		  gtk_text_insert(GTK_TEXT(Editor),
				  NULL, &(Prefs.Colors.L[Lg].Comment),
				  NULL,
				  buffer + i, -1);
		  i = buflen;
		}
	    }
	  else
	    {
	      if ((gint)NULL !=
		  (j = (gint)strstr(buffer + i +
				    strlen(Prefs.L[Lg].BlockCommentOn),
				    Prefs.L[Lg].BlockCommentOff)))
		{
		  j = j + strlen(Prefs.L[Lg].BlockCommentOff) - (gint)buffer;
		  gtk_text_insert(GTK_TEXT(Editor),
				  NULL, &(Prefs.Colors.L[Lg].Comment),
				  NULL,
				  buffer + i, j - i);
		  i = j;
		}
	      else
		{
		  gtk_text_insert(GTK_TEXT(Editor),
				  NULL, &(Prefs.Colors.L[Lg].Comment),
				  NULL,
				  buffer + i, -1);
		  i = buflen;
		}
	    }
	}
      /*** Detection of BlockCommentOnAlt */
      else if ((Prefs.L[Lg].BlockCommentOnAlt != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].BlockCommentOnAlt,
			strlen(Prefs.L[Lg].BlockCommentOnAlt)))
	{
	  if (Prefs.L[Lg].BlockCommentOffAlt == NULL)
	    {
	      if ((j = (gint)strstr(buffer + i, "\n")) != (gint)NULL)
		{
		  j = j + 1 - (gint)buffer;
		  gtk_text_insert(GTK_TEXT(Editor), NULL,
				  &(Prefs.Colors.L[Lg].CommentAlt), NULL,
				  buffer + i, j - i);
		  i = j;
		}
	      else
		{
		  gtk_text_insert(GTK_TEXT(Editor),
				  NULL,
				  &(Prefs.Colors.L[Lg].CommentAlt),
				  NULL,
				  buffer + i, -1);
		  i = buflen;
		}
	    }
	  else
	    {
	      if ((gint)NULL !=
		  (j = (gint)strstr(buffer + i,
				    Prefs.L[Lg].BlockCommentOffAlt)))
		{
		  j = j + strlen(Prefs.L[Lg].BlockCommentOffAlt) -
		    (gint)buffer;
		  gtk_text_insert(GTK_TEXT(Editor), NULL,
				  &(Prefs.Colors.L[Lg].CommentAlt), NULL,
				  buffer + i, j - i);
		  i = j;
		}
	      else
		{
		  gtk_text_insert(GTK_TEXT(Editor),
				  NULL,
				  &(Prefs.Colors.L[Lg].CommentAlt),
				  NULL,
				  buffer + i, -1);
		  i = buflen;
		}
	    }
	}
      /*** Detection of String0 */
      else if (Prefs.L[Lg].HaveString && 
	       (Prefs.L[Lg].StringChar0 != '\0') &&
	       (buffer[i] == Prefs.L[Lg].StringChar0))
	{
	  j = i + 1;
	  while ((j < buflen) && (buffer[j] != Prefs.L[Lg].StringChar0))
	    {
	      if (buffer[j] == Prefs.L[Lg].EscapeChar)
		j += 2;
	      else
		j++;
	    }
	  if (j >= buflen)
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].String0),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].String0),
			      NULL,
			      buffer + i, j + 1 - i);
	      i = j + 1;
	    }
	}
      /*** Detection of String1 */
      else if (Prefs.L[Lg].HaveString && 
	       (Prefs.L[Lg].StringChar1 != '\0') &&
	       (buffer[i] == Prefs.L[Lg].StringChar1))
	{
	  j = i + 1;
	  while ((j < buflen) && (buffer[j] != Prefs.L[Lg].StringChar1))
	    {
	      if (buffer[j] == Prefs.L[Lg].EscapeChar)
		j += 2;
	      else
		j++;
	    }
	  if (j >= buflen)
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].String1),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].String1),
			      NULL,
			      buffer + i, j + 1 - i);
	      i = j + 1;
	    }
	}
      /*** Detection of LineComment */
      else if ((Prefs.L[Lg].LineComment != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].LineComment,
			strlen(Prefs.L[Lg].LineComment)))
	{
	  j = i + 1;
	  while ((j < buflen) && (buffer[j] != '\n'))
	      j++;
	  if (j < buflen)
	    {
	      j++;
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].Comment),
			      NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].Comment),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
      /*** Detection of LineCommentAlt */
      else if ((Prefs.L[Lg].LineCommentAlt != NULL) &&
	       !strncmp(buffer + i, Prefs.L[Lg].LineCommentAlt,
			strlen(Prefs.L[Lg].LineCommentAlt)))
	{
	  j = i + 1;
	  while ((j < buflen) && (buffer[j] != '\n'))
	      j++;
	  if (j < buflen)
	    {
	      j++;
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].CommentAlt),
			      NULL,
			      buffer + i, j - i);
	      i = j;
	    }
	  else
	    {
	      gtk_text_insert(GTK_TEXT(Editor),
			      NULL,
			      &(Prefs.Colors.L[Lg].CommentAlt),
			      NULL,
			      buffer + i, -1);
	      i = buflen;
	    }
	}
      /*** Detection of Number */
      else if (isdigit(buffer[i]))
	{
	  j = strcspn(buffer + i, Prefs.L[Lg].Delimiters);
	  gtk_text_insert(GTK_TEXT(Editor),
			  NULL, &(Prefs.Colors.L[Lg].Number), NULL,
			  buffer + i, j);
	  i += j;
	}
      /*** Detection of: Keywords or Delimiters or nothing */
      else
	{
	  t_match	match;

	  /* Detection of Keywords or Delimiters */
	  if ((word_to_correct == i) &&
	      Prefs.L[Lg].IsCaseSensitive && Prefs.AutoCorrection )
	    matching_keyword_case_ext(Lg, buffer + i, &match);
	  else
	    matching_keyword(Lg, buffer + i, &match);
	  /* If length is != 0, a matching keyword has been found */
	  if (match.length)
	    {
#ifdef DEBUG_SYHI
	      g_print(__FILE__": "__func__"(): Insertion of Keyword or "
		      "Delimiters reconized, i = %i\n", i);
#endif
	      if (Prefs.L[Lg].IsCaseSensitive && Prefs.AutoCorrection &&
		  (word_to_correct == i))
		gtk_text_insert(GTK_TEXT(Editor), NULL, &(Prefs.Colors.L[Lg]\
		.C[match.color]), NULL, match.keyword, match.length),
		word_to_correct = -1;
	      else
		gtk_text_insert(GTK_TEXT(Editor), NULL, &(Prefs.Colors.L[Lg]\
		.C[match.color]), NULL, buffer + i, match.length);
	      i += match.length;
	    }
	  /* If length is 0, nothing has been detected */
	  else
	    {
#ifdef DEBUG_SYHI
	      g_print(__FILE__": "__func__"(): Insertion of unknown string, "
		      "i = %i\n", i);
#endif
	      /*
	      ** NOTE_CHANGE: Changed on 07-oct-2000
	      ** if ((j = strcspn(buffer + i, Prefs.L[Lg].Delimiters)) == 0)
	      **  j = 1;
	      */
	      j = 0;
	      while (!Prefs.L[Lg].IsADelimiter[(guchar)buffer[i + j]])
		j++;
	      j = (j) ? (j) : (1);
	      /*
	      ** NOTE_CHANGE: End
	      */
	      gtk_text_insert(GTK_TEXT(Editor), NULL,NULL,NULL, buffer + i, j);
	      i += j;
	    }
	}
    } //while (i < buflen && ...)
  if (i < buflen)
      /* So we have reached and/or gone beyond CursorPosition +
	 Prefs.SyhiUpdate */
      gtk_text_insert(GTK_TEXT(Editor), NULL,
		      NULL, NULL,
		      buffer + i, -1);
#ifdef DEBUG_AUTO_INDENT
  if (Prefs.AutoIndentation)
      g_print(__FILE__": "__func__"(): * AutoIndentation activated\n");
  else
      g_print(__FILE__": "__func__"(): * AutoIndentation desactivated\n");
  g_print(__FILE__": "__func__"(): refresh_from = %i\n", refresh_from);
#endif
  /* If a new line has just been inserted, proceed for auto-indentation */
  if (Prefs.AutoIndentation && enter_pressed)
    {
      gboolean		indent;
      gboolean		unindent;
      gint		first;
      gint		first_nospc;
      gint		last_nospc;

#ifdef DEBUG_AUTO_INDENT
      g_print(__FILE__": "__func__"(): AutoIndentation starting\n");
#endif
      enter_pressed = FALSE;
      /* NOTE_REMINDER: buffer[refresh_from] is the first char of
	 the previous line */
      first = refresh_from;
      /* Here buffer[first] is the first char of previous line */
#ifdef DEBUG_AUTO_INDENT
      g_print(__FILE__": "__func__"(): buffer[first]: [%c,%c,%c,%c]\n",
	      PRINTC(buffer[first + 0]),
	      PRINTC(buffer[first + 1]),
	      PRINTC(buffer[first + 2]),
	      PRINTC(buffer[first + 3]));
#endif
      first_nospc = first; 
      while ((buffer[first_nospc] == ' ') || (buffer[first_nospc] == '\t'))
	first_nospc++;
      /* Here buffer[first_nospc] is the 1st non-space char of previous line */
#ifdef DEBUG_AUTO_INDENT
      g_print(__FILE__": "__func__"(): buffer[first_nospc]: [%c,%c,%c,%c]\n",
	      PRINTC(buffer[first_nospc + 0]),
	      PRINTC(buffer[first_nospc + 1]),
	      PRINTC(buffer[first_nospc + 2]),
	      PRINTC(buffer[first_nospc + 3]));
#endif
      last_nospc = (gint)strchr(buffer + first_nospc, '\n') - 1 - (gint)buffer;
      while ((last_nospc > first_nospc) &&
	     ((buffer[last_nospc] == ' ') || (buffer[last_nospc] == '\t')))
	last_nospc--;
      last_nospc = (last_nospc < first_nospc) ? first_nospc : last_nospc;
      /* Here buffer[last_nospc] is the last non-space char of previous line */
#ifdef DEBUG_AUTO_INDENT
      g_print(__FILE__": "__func__"(): buffer[last_nospc]: [%c,%c,%c,%c]\n",
	      PRINTC(buffer[last_nospc + 0]),
	      PRINTC(buffer[last_nospc + 1]),
	      PRINTC(buffer[last_nospc + 2]),
	      PRINTC(buffer[last_nospc + 3]));
#endif
      indent = unindent = FALSE;
      if (Prefs.L[Lg].UnindentString &&
	  (last_nospc + 1 - first_nospc >=
	   (gint)strlen(Prefs.L[Lg].UnindentString)) &&
	  !strncmp(buffer + first_nospc, Prefs.L[Lg].UnindentString,
		   strlen(Prefs.L[Lg].UnindentString)))
	/*** If the first non-space chars of previous line are the
	     UnindentString... auto-unindentation */
	unindent = TRUE;
      if (Prefs.L[Lg].IndentString &&
	  (last_nospc + 1 - first_nospc >=
	   (gint)strlen(Prefs.L[Lg].IndentString)) &&
	  !strncmp(buffer + last_nospc + 1 - strlen(Prefs.L[Lg].IndentString),
		   Prefs.L[Lg].IndentString, strlen(Prefs.L[Lg].IndentString)))
	/*** If the last non-space chars of previous line are the
	     IndentString... auto-indentation */
	indent = TRUE;
      if (unindent)
	{
	  gint		indent_space;

#ifdef DEBUG_AUTO_INDENT
	  g_print(__FILE__": "__func__"(): --- Unindentation\n");
#endif /*DEBUG_AUTO_INDENT*/
	  indent_space = 0;
	  for (j = first; j < first_nospc; j++)
	    if (buffer[j] == ' ')
	      indent_space++;
	    else
	      indent_space += Prefs.TabSize - (indent_space % Prefs.TabSize);
	  /* Here, indent_space is the size in chars of indentation of
	     previous line */
#ifdef DEBUG_AUTO_INDENT
	  g_print("Indentation space of previous line = %i\n", indent_space);
#endif /*DEBUG_AUTO_INDENT*/
	  if (indent_space > 0)
	    {
	      gint	indent_chars;
	      gint	indent_space_from_bol; /* bol = Beginning Of Line */

	      if (!(indent_space % Prefs.TabSize))
		indent_space -= Prefs.TabSize;
	      else
		indent_space -= indent_space % Prefs.TabSize;
#ifdef DEBUG_AUTO_INDENT
	      g_print("Indentation space required for previous & current "
		      "line = %i\n", indent_space);
#endif /*DEBUG_AUTO_INDENT*/
	      /* Here, indent_space is the size in spaces of indentation that
		 current line should have. Now, indent_chars should be the
		 size in reals chars (tab, space...) of indentation that
		 current line should have. */
	      indent_space_from_bol = 0;
	      for (i = first; indent_space > 0; i++)
		if (buffer[i] == ' ')
		  {
		    indent_space--;
		    indent_space_from_bol++;
		  }
		else
		  {
		    gint isfb_tmp; /* Indent Space From Bol TeMP */

		    isfb_tmp = indent_space_from_bol;
		    indent_space -= Prefs.TabSize - isfb_tmp % Prefs.TabSize;
		    indent_space_from_bol += isfb_tmp % Prefs.TabSize;
		  }
	      if (indent_space != 0)
		g_print(__FILE__": "__func__"(): Fatal error, indent_space "
			"!= 0 (== %i)\nYou should save all your documents "
			"and quit Beaver immediately !\n", indent_space);
	      indent_chars = i - first;
	      /* the (indent_chars)th first chars of previous line should
		 be used to indent the new line */
#ifdef DEBUG_AUTO_INDENT
	      g_print("Struct of indentation of prev. & act. lines: "
		      "[first=%i; first+indent_chars-1=%i] "
		      "(indent_chars = %i): \"",
		      first, first + indent_chars - 1, indent_chars);
	      for (i = first; i < first + indent_chars; i++)
		if (buffer[i] == ' ')
		  g_print(".");
		else
		  g_print("-");
	      g_print("\"\n");
	      g_print("first_nospc = %i; last_nospc = %i\n"
		      "refresh_from = %i =?= CursorPosition = %i\n",
		      first_nospc, last_nospc, refresh_from, CursorPosition);
	      g_print("Deleting [%i; %i] (%i chars)...\n",
		      first + indent_chars, first_nospc - 1,
		      first_nospc - (first + indent_chars));
#endif /*DEBUG_AUTO_INDENT*/
	      gtk_editable_delete_text(GTK_EDITABLE(Editor),
				       first + indent_chars,
				       first_nospc);
	      CursorPosition -= first_nospc - (first + indent_chars);
	      first_nospc -= first_nospc - (first + indent_chars);
#ifdef DEBUG_AUTO_INDENT
	      g_print("Now, CursorPosition = %i\n", CursorPosition);
#endif /*DEBUG_AUTO_INDENT*/
	      gtk_text_set_point(GTK_TEXT(Editor), CursorPosition);
	      if (indent_chars > 0)
		{
#ifdef DEBUG_AUTO_INDENT
		  g_print("New line had to be indented of %i chars\n",
			  indent_chars);
#endif /*DEBUG_AUTO_INDENT*/
		  gtk_text_insert(GTK_TEXT(Editor), NULL,
				  NULL, NULL,
				  buffer + first, indent_chars);
		  CursorPosition += indent_chars;
#ifdef DEBUG_AUTO_INDENT
		  g_print("Now, CursorPosition = %i\n", CursorPosition);
#endif /*DEBUG_AUTO_INDENT*/
		}
	    } /* if (indent_space > 0) */
	} /*** Unindentation */
      if (indent)
	{
#ifdef DEBUG_AUTO_INDENT
	  g_print(__FILE__": "__func__"(): --- Indentation\n");
#endif /*DEBUG_AUTO_INDENT*/
	  indent = TRUE;
	  gtk_text_set_point(GTK_TEXT(Editor), CursorPosition);
#ifdef DEBUG_AUTO_INDENT
	  g_print("Now, CursorPosition = %i\n", CursorPosition);
#endif /*DEBUG_AUTO_INDENT*/
	  if (!unindent)
	    gtk_text_insert(GTK_TEXT(Editor), NULL,
			    NULL, NULL,
			    buffer + first, first_nospc - first);
	  gtk_text_insert(GTK_TEXT(Editor), NULL,
			  NULL, NULL,
			  "\t", 1);
	  if (!unindent)
	    CursorPosition += first_nospc - first + 1;
	  else
	    CursorPosition++;
	} /*** Indentation */
      if (!indent && !unindent)
	{
	  /*** Else keep same indentation as previous line */
#ifdef DEBUG_AUTO_INDENT
	  g_print(__FILE__": "__func__"(): --- Same indentation\n");
#endif /*DEBUG_AUTO_INDENT*/
	  gtk_text_set_point(GTK_TEXT(Editor), CursorPosition);
	  gtk_text_insert(GTK_TEXT(Editor), NULL,
			  NULL, NULL,
			  buffer + first, first_nospc - first);
	  CursorPosition += first_nospc - first;
	} /*** Same indentation */
#ifdef DEBUG_SYHI
      g_print(__FILE__": "__func__"(): End of AutoIndentation\n");
#endif
    }
  /*
  ** End of Auto indentation
  */
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): All text has been inserted, freeing...\n");
#endif
  g_free(buffer);
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Thawing Editor...\n");
#endif
  /* gtk_text_thaw() must be placed before gtk_editable_set_position(), */
  /* If not, this causes a seg fault !! */
  gtk_text_thaw(GTK_TEXT(Editor));
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Setting position...\n");
#endif
  /* I don't know why, but sometimes the gtk_text_set_point() does not work...
     gtk_text_set_point(GTK_TEXT(Editor), CursorPosition); */
  gtk_editable_set_position(GTK_EDITABLE(Editor), CursorPosition);
#ifdef DEBUG_SYHI
  g_print(__FILE__": "__func__"(): Connecting_after...\n");
#endif
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectChgAft) =
    gtk_signal_connect_after(GTK_OBJECT(Editor), "changed",
			     (GtkSignalFunc)reconnect_sig,
			     NULL);
  refresh_from = 0;
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return TRUE;
}

/*
** Callback fcn for changed signal ("connected_after" the refresh_syhi)
**
** Parameters :
**  See a gtk reference manual
**
** Return values :
**  TRUE		Always
**  FALSE		Never
*/

gboolean		reconnect_sig(GtkEditable *Editor, gpointer data)
{
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  (void)data;
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectIns) =
    gtk_signal_connect(GTK_OBJECT(Editor), "insert_text",
		       (GtkSignalFunc)text_has_been_inserted,
		       NULL);
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectDel) =
    gtk_signal_connect(GTK_OBJECT(Editor), "delete_text",
		       (GtkSignalFunc)text_has_been_deleted,
		       NULL);
  //NOTE_WARNING : the callback fcn refresh_syhi() needs the "changed"
  //signal to be produced after "delete_text" ans "insert_text" signals 
  FPROPS(gtk_notebook_get_current_page(GTK_NOTEBOOK(MainNotebook)),
	 WidgetInfo.SigConnectChg) =
    gtk_signal_connect(GTK_OBJECT(Editor), "changed",
		       (GtkSignalFunc)refresh_syhi,
		       NULL);
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
  return TRUE;
}

/*
** This fcn find the longest keyword that match the string passed in
** arguments, taking into account IsHTML value.
**
** Only .lenght and .color field are filled
**
** Parameters :
**  Lg			The language of the string
**  str			The string this fcn try to match with a keyword
**  match		Structure representing the mathcing keyword filled
**			 by this function
**
** Return values :
**  void
*/
void			matching_keyword(gint Lg,
					 const guchar *str,
					 t_match *match)
{
  int			w_length;
  t_wtree		*walk;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  match->length = w_length = 0;
  walk = Prefs.L[Lg].wtree_root;
  while (str[w_length] && (walk = walk->letter[str[w_length++]]))
    if (walk->is_eow &&
	(!str[w_length] ||
	 (Prefs.L[Lg].IsADelimiter[str[w_length]] ||
	  (Prefs.L[Lg].IsHTML && str[w_length - 1] == '>'))))
      {
	match->length = w_length;
	match->color = walk->color_plus_1 - 1;
      }
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
}

/*
** This func find the longest keyword that match the beginning of the
** string passed in arguments, regarding the value of IsHTML.
**
** It is said ``case extended'' because it fills 2 more fields of
**  match than matching_keyword(): .keyword and .good_case. This allows
**  to match some keywords even if they do not have the same case that
**  the keyword in the wordfile
**
**  This func returns its results in a struct s_match where:
**  - match.length is 0 if there are no keywords matching the string, or
**    match.lenght contains the length of match.keyword (a null-terminated
**    string matching the string, need to be freed ?). The case of this
**    match.keyword is always the one of the keyword written in the wordfile
**  - match.color is the number of the group color of the matching keyword
**
** Parameters :
**  Lg			The language of the string
**  str			The string this fcn try to match with a keyword
**  match		Pointer on the struct that this func filled
**
** Return value :
**  void
*/
void			matching_keyword_case_ext(gint Lg,
						  const guchar *str,
						  t_match *match)
{
  gchar			*search_ptr;
  gchar			*tmp_ptr;
  gint			color;
  gint			len_search_ptr;
  gint			cmp;

#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): Begin\n");
#endif
  match->length = 0;
  for (color = 0; color < MAX_COL; color++)
    {
      if (Prefs.L[Lg].C[color].Keywords == NULL)
	continue;
      search_ptr = Prefs.L[Lg].C[color].Keywords + 1;
      /* NOTE_REMARK: Keywords are separated by spaces: " k1 k2 k3 " */
      while (search_ptr != NULL)
	{
	  tmp_ptr = strchr(search_ptr, ' ');
	  *tmp_ptr = '\0';
	  len_search_ptr = tmp_ptr - search_ptr;
	  if ((CHAR_CASE_CMP(*str, *search_ptr)) &&
	      (len_search_ptr > match->length) &&
	      (!(cmp = strncmp(str, search_ptr, len_search_ptr)) ||
	       !strncasecmp(str, search_ptr, len_search_ptr)) &&
	      (Prefs.L[Lg].IsADelimiter[str[len_search_ptr]] ||
	       (Prefs.L[Lg].IsHTML && str[len_search_ptr - 1] == '>')))
	    {
	      match->color = color;
	      match->length = len_search_ptr;
	      match->keyword = search_ptr;
	      match->good_case = !cmp ? TRUE : FALSE;
	    }
	  *tmp_ptr = ' ';
	  search_ptr = tmp_ptr + 1;
	  if (*search_ptr == '\0')
	    search_ptr = NULL;
	}
    }
  /*    - match->length == 0
     or - match->length != 0 and match->keyword points to a matching keyword */
#ifdef DEBUG_FCN
  g_print(__FILE__": "__func__"(): End\n");
#endif
}
